/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */

#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "adf.h"
#include "aps.h"
#include "box.h"
#include "browse.h"
#include "commands.h"
#include "dis_if.h"
#include "drone.h"
#include "flaps.h"
#include "gear.h"
#include "hud.h"
#include "hsi.h"
#include "instruments.h"
#include "joystick.h"
#include "mouse.h"
#include "navaid.h"
#include "planes.h"
#include "players.h"
#include "pm.h"
#include "prompt.h"
#include "radar.h"
#include "render.h"
#include "terminal.h"
#include "update.h"
#include "weapon.h"
#include "windows.h"

#define MAX_MAPPED_STRING_LEN	20
#define MAX_POPUP_STRING_LEN	40

#define FLIGHT_STATE_FILE "./acm-flight-state"

/*
	To prevent dangerous typos from causing unfortunate consequences,
	several events require double typing with a couple seconds to be
	activated: Quit (XK_Q), change stick (XK_m), and calibrate joystick
	(XK_K).  These variables hold the time of the previous key event.
*/
static double pressed_XK_Q_time = -5.0;
static double pressed_XK_m_time = -5.0;
static double pressed_XK_K_time = -5.0;

struct		keymasks {
char		b1;
char		b2;
char		b3;
char		b4;
char		b5;
char		sm;
char		lm;
char		cm;
char		m1;
char		m2;
char		m3;
char		m4;
char		m5;
char		n;
} static	kmsks = {	'.', '.', '.', '.', '.',
				'.', '.', '.',
				'.', '.', '.', '.', '.', '\0'};

KeySym		keysym;
XComposeStatus	compose;
char		buffer[MAX_MAPPED_STRING_LEN];
int		buflen = MAX_MAPPED_STRING_LEN;
KeySym		keysym;
char *		symstr;
char	  	viewermsg[121];
const char	evtfmt[] = "%s %c%c %16s\n";
const char	actfmt[] = "%8.2f         %16s %s\n";


static _BOOL
getEventInfo(craft * c, viewer * u, XEvent * ev) 
{

	char	dtype, etype;

	dtype = ev->type == KeyPress || ev->type == KeyRelease  ? 'K' : 'M';
	etype = ev->type == KeyPress || ev->type == ButtonPress ? 'V' : '^';
		
	if (dtype == 'K') {
		if ( ! u->hasFocus )  {
			fprintf(stderr, "%s\n", "Keyboard event ignored -- no focus");
			return FALSE;
		}

        	kmsks.b1 = ev->xkey.state & Button1Mask ? '1' : '.';
        	kmsks.b2 = ev->xkey.state & Button1Mask ? '2' : '.';
        	kmsks.b3 = ev->xkey.state & Button1Mask ? '3' : '.';
        	kmsks.b4 = ev->xkey.state & Button1Mask ? '4' : '.';
        	kmsks.b5 = ev->xkey.state & Button1Mask ? '5' : '.';
        	kmsks.sm = ev->xkey.state & ShiftMask   ? 'S' : '.';
        	kmsks.lm = ev->xkey.state & LockMask    ? 'L' : '.';
        	kmsks.cm = ev->xkey.state & ControlMask ? 'C' : '.';
        	kmsks.m1 = ev->xkey.state & Mod1Mask    ? '1' : '.';
        	kmsks.m2 = ev->xkey.state & Mod1Mask    ? '2' : '.';
        	kmsks.m3 = ev->xkey.state & Mod1Mask    ? '3' : '.';
        	kmsks.m4 = ev->xkey.state & Mod1Mask    ? '4' : '.';
        	kmsks.m5 = ev->xkey.state & Mod1Mask    ? '5' : '.';

		(void) XLookupString((XKeyEvent *) ev,
					buffer, buflen, &keysym, &compose);

 		if ((symstr = (char *) XKeysymToString(keysym)) == NoSymbol)
			symstr = &"[NoSymbol]";
	} else {
		if ( ! u->hasFocus | ! u->hasComm ) {
			fprintf(stderr, "%s\n",
				"Non-keyboard event ignored (no focus and/or commands are disabled)");
			return FALSE;
		}

	        kmsks.b1 = ev->xbutton.state & Button1Mask ? '1' : '.';
        	kmsks.b2 = ev->xbutton.state & Button1Mask ? '2' : '.';
        	kmsks.b3 = ev->xbutton.state & Button1Mask ? '3' : '.';
        	kmsks.b4 = ev->xbutton.state & Button1Mask ? '4' : '.';
        	kmsks.b5 = ev->xbutton.state & Button1Mask ? '5' : '.';
        	kmsks.sm = ev->xbutton.state & ShiftMask   ? 'S' : '.';
        	kmsks.lm = ev->xbutton.state & LockMask    ? 'L' : '.';
        	kmsks.cm = ev->xbutton.state & ControlMask ? 'C' : '.';
        	kmsks.m1 = ev->xbutton.state & Mod1Mask    ? '1' : '.';
        	kmsks.m2 = ev->xbutton.state & Mod1Mask    ? '2' : '.';
        	kmsks.m3 = ev->xbutton.state & Mod1Mask    ? '3' : '.';
        	kmsks.m4 = ev->xbutton.state & Mod1Mask    ? '4' : '.';
        	kmsks.m5 = ev->xbutton.state & Mod1Mask    ? '5' : '.';

		switch (ev->xbutton.button) {
		case Button1: symstr = &"1" ; break;
		case Button2: symstr = &"2" ; break;
		case Button3: symstr = &"3" ; break;
		case Button4: symstr = &"4" ; break;
		case Button5: symstr = &"5" ; break;
		}
	}

	fprintf(stderr, evtfmt, (char *) &kmsks, dtype, etype, symstr);

	return TRUE;
}


static void
updateHUDColor(viewer * u)
{
	HUDColor = HUDColors[u->hColor * 4 + u->hBright];
	HUDPixel = HUDColor->cIndex;
	u->hdgScale.pixel = u->v->w->pixel[HUDPixel];
	u->altScale.pixel = u->v->w->pixel[HUDPixel];
	u->velScale.pixel = u->v->w->pixel[HUDPixel];
}


static _BOOL
doButtonEvent(craft * c, viewer * u, XEvent * ev)
{
	if (! getEventInfo(c, u, ev)) {
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--button event ignored");
		return FALSE;
	}

	if (ev->xbutton.button == Button1) {
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    selectCockpitItem");
		selectCockpitItem( c, u, 
						   ev->xbutton.x, 
						   ev->xbutton.y, 
						   ev->xbutton.time );
	}
	else if (ev->xbutton.button == Button2) {
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    fireWeapon");
		fireWeapon(c);
	}
	else if (ev->xbutton.button == Button3) {
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    selectWeapon");
		selectWeapon(c);
	}
	return TRUE;
}


static _BOOL
doButtonReleaseEvent(craft * c, viewer * u, XEvent * ev, int player)
{

	if (! getEventInfo(c, u, ev))  {
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--button release event ignored");
		return FALSE;
	}

	if (ev->xbutton.button == Button2)
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    ceaseFireWeapon");
		ceaseFireWeapon(c);
	return TRUE;
}


#define SWITCH_TO_NEXT_PLAYER_ENABLED

#ifdef SWITCH_TO_NEXT_PLAYER_ENABLED

static void
switchToNextPlayer(viewer * v)
{
	int    i;
	craft *p;

	for ((i = 0, p = ptbl); i < MAXPLAYERS; (++i, ++p)) {
		if ( ((p->type == CT_DRONE) || (p->type == CT_PLANE))
		&& p->vl == NULL ) {
			break;
		}
	}

	if ( i == MAXPLAYERS ) {
		prompt_viewer_print(v, "No other plane available");
		return;
	}

	v->c->vl = NULL; /* FIXME: vl might contain several viewers */
	v->c = p;
	p->vl = v;
	/****
	if ( p->type == CT_DRONE )
		v->hasComm = FALSE;
	else
		v->hasComm = TRUE;
	****/
}

#endif


static _BOOL
doKeyRelease(craft * c, viewer * u, XKeyEvent * ev)
/*
	Handle key release event. Return TRUE is player gets killed or player
	asked to terminate the program.
*/
{

	if (! getEventInfo(c, u, (XEvent *) ev)) {
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--key release event ignored");
		return FALSE;
	}
	if (keysym == XK_v)
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    ceaseFireWeapon");
		ceaseFireWeapon(c);
	return FALSE;
}


static _BOOL
doKeyEvent(craft * c, viewer * u, XKeyEvent * ev)
/*
	Handle key press event. Return TRUE is player gets killed or player
	asked to terminate the program.
*/
{

#ifdef SPECIAL_KEYS
	FILE     *fp;
	craft     pentry;
#endif

	if (! getEventInfo(c, u, (XEvent *) ev)) {
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--key press event ignored");
		return FALSE;
	}

	if ( ! u->hasComm ){
		if ( keysym == XK_d ) {
			fprintf(stderr, actfmt, curTime, symstr,
				"attaching  ACM back to X inputs");
			u->hasComm = TRUE;
		} else {
			fprintf(stderr, "%s\n",
				"Non-'d' key press ignored (commands are disabled)");
		}
		return FALSE;
	}

	if ( terminal_enabled(u) && (keysym != XK_F1) ){
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    terminal_input_char");
		terminal_input_char(u, buffer[0]);
		{
			char line[100];
			if( terminal_read_string(u, line, 100) ) {
				fprintf(stderr, actfmt, curTime, symstr,
					"calling    commands_execute");
				commands_execute(u, line);
			}
		}
		return FALSE;
	}

	switch (keysym) {

	case XK_F1:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    terminal_toggle");
		terminal_toggle(u);
		break;

#ifdef SWITCH_TO_NEXT_PLAYER_ENABLED
	case XK_Tab:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    switchToNextPlayer");
		switchToNextPlayer(u);
		break;
#endif

	case XK_d:
		fprintf(stderr, actfmt, curTime, symstr,
			"detaching  ACM from X inputs");
		u->hasComm = FALSE;
		break;

	case XK_K:
		if ( pressed_XK_K_time + 2.0 >= curTime ) {
			fprintf(stderr, actfmt, curTime, symstr,
				"calling    CalibrateJoystick");
			CalibrateJoystick();
			return TRUE;
		}
		else {
			fprintf(stderr, actfmt, curTime, symstr,
				"arming     CalibrateJoystick");
			pressed_XK_K_time = curTime;
			prompt_viewer_print(u, "Press SHIFT-K again within 2 sec. to calibrate the joystick");
		}
		break;

	case XK_m:
		if ( pressed_XK_m_time + 2.0 >= curTime ){
		    if (u->noMorJ) {
			fprintf(stderr, actfmt, curTime, symstr,
				"toggling   stick to mouse");
			u->noMorJ = FALSE;
			prompt_viewer_print(u, "Mouse pitch and roll now ACTIVE");
		    } else {
			fprintf(stderr, actfmt, curTime, symstr,
				"toggling   stick to keyboard");
			u->noMorJ = TRUE;
			prompt_viewer_print(u, "Mouse pitch and roll now INACTIVE");
		    }
			return TRUE;
		}
		else {
			fprintf(stderr, actfmt, curTime, symstr,
				"arming     stick toggle");
			pressed_XK_m_time = curTime;
			prompt_viewer_print(u, "Press 'm' again within 2 sec. to switch the stick control");
		}
		break;

	case XK_U:
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   keyboard cursor display");
		u->showMark ^= 1;
		break;
	case XK_v:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    fireWeapon");
		fireWeapon(c);
		break;

	case XK_V:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    selectWeapon");
		selectWeapon(c);
		break;

	case XK_G:
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    ground color");
		if ( gtblCount < 2 )
			prompt_viewer_print(u, "Only one GROUNDC defined");
		else {
    			if ( gtblIndex == gtblCount - 1 )
				gtblIndex = 0;
			else    gtblIndex++;
    			groundColor = gtbl[gtblIndex].vcptr;
			sprintf(viewermsg, "groundColor is now %s", gtbl[gtblIndex].name);
			prompt_viewer_print(u, viewermsg);
		}
		break;

	case XK_C:
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    HUD color");
		u->hColor = u->hColor == 3 ? 0 : u->hColor + 1;
		updateHUDColor(u);
		break;

	case XK_B:
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    HUD brightness");
		u->hBright = u->hBright == 3 ? 0 : u->hBright + 1;
		updateHUDColor(u);
		break;

	case XK_H:
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   HUD display");
		windows_set_layout(c, u, u->w->width, u->w->height, !u->hud_mode);
		break;

	case XK_A:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_toggle");
		aps_ap_toggle(c);
		break;

	case XK_5:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(-1.0)");
		aps_ap_adj_vs(c, -1.0);
		break;

	case XK_6:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(+1.0)");
		aps_ap_adj_vs(c, +1.0);
		break;

	case XK_percent:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(-100.0)");
		aps_ap_adj_vs(c, -100.0);
		break;

	case XK_asciicircum:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(+100.0)");
		aps_ap_adj_vs(c, +100.0);
		break;

	case XK_Z:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_toggle");
		aps_ap_toggle(c);
		if( aps_ap_enabled(c) )
			aps_ap_set_vs(c, 0.0);
		break;

	case XK_X:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ac_toggle");
		aps_ac_toggle(c);
		break;

	case XK_T:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_at_toggle");
		aps_at_toggle(c);
		break;

	case XK_N:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_an_toggle");
		aps_an_toggle(c);
		break;

	case XK_L:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_al_toggle");
		aps_al_toggle(c);
		break;

	case XK_parenleft:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_bank_max_dec");
		aps_bank_max_dec(c);
		break;

	case XK_parenright:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_bank_max_inc");
		aps_bank_max_inc(c);
		break;

	case XK_M:
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   magnetic/true");
		/* Toggle MH<->TH display */
		c->showMag = ! c->showMag;
		break;

	case XK_less:
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 3.0 left");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(-3.0));
		break;

	case XK_greater:
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 3.0 right");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(3.0));
		break;

	case XK_comma:
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 1.5 left");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(-1.5));
		break;

	case XK_period:
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 1.5 right");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(1.5));
		break;

	case XK_bar:
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for wings level");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(0.0));
		break;

	case XK_slash:
		fprintf(stderr, actfmt, curTime, symstr,
			"disabing   AW and AN and AL");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_disable(c);
		break;

	case XK_Up:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for ^");
		getStick(u->c, u, 1);
		break;

	case XK_Down:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for V");
		getStick(u->c, u, 2);
		break;

	case XK_Left:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for <");
		getStick(u->c, u, 3);
		break;

	case XK_Right:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for >");
		getStick(u->c, u, 4);
		break;

#ifdef sun
	case XK_R2:
#else
	case XK_Page_Up:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing keystick sensitivity");
		if (u->stickM == KEYSTICK_LINEAR)
			u->stickS++;
		else	u->stickS = u->stickS << 1;
		if ( u->stickS > 128 )
			u->stickS = 128;
		break;

#ifdef sun
	case XK_R5:
#else
	case XK_Page_Down:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing keystick sensitivity");
		if (u->stickM == KEYSTICK_LINEAR)
			u->stickS--;
		else	u->stickS = u->stickS >> 1;
		if ( u->stickS < 1 )
			u->stickS = 1;
		break;

	case XK_End:
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    keystick mode");
		if (u->stickM == KEYSTICK_LINEAR)
			u->stickM = KEYSTICK_DOUBLE;
		else	u->stickM = KEYSTICK_LINEAR;
		break;

	case XK_z:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing rudderComm (more left)");
		c->rudderComm = c->rudderComm + 0.01;
		if ( c->rudderComm > 1.0 )  c->rudderComm = 1.0;
		break;

	case XK_c:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing rudderComm (more right)");
		c->rudderComm = c->rudderComm - 0.01;
		if ( c->rudderComm < -1.0 )  c->rudderComm = -1.0;
		break;

	case XK_x:
		fprintf(stderr, actfmt, curTime, symstr,
			"zeroing    rudderComm (center)");
		c->rudderComm = 0.0;
		break;

#ifdef sun
	case XK_R7:
#else
	case XK_Home:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"cancling   all APS and trim settings");
		aps_off(c);
		c->SeTrim = 0.0;
		c->SaTrim = 0.0;
		break;

	case XK_j:
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    elevator trim to current stick");
		/* Adjust Elevator Trim */
		c->SeTrim -= c->pitchComm;
		break;

	case XK_y:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_up");
		flaps_up(c);
		break;

	case XK_h:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_down");
		flaps_down(c);
		break;

	case XK_w:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_speed_brakes_retract");
		flaps_speed_brakes_retract(c);
		break;

	case XK_s:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_speed_brakes_extend");
		flaps_speed_brakes_extend(c);
		break;

	case XK_t:
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    timer");
		if( u->hud_mode )
			hud_timer_toggle(u);
		else
			inst_timer_toggle(u);
		break;

	case XK_1:
/*	case XK_Next:	*/

		fprintf(stderr, actfmt, curTime, symstr,
			"setting    throttle to idle (disable all APS except AC)");
		aps_al_disable(c);
		aps_an_disable(c);
		aps_ap_disable(c);
		aps_at_disable(c);
		aps_aw_disable(c);

		c->throttleComm = 8192;
		break;

	case XK_2:
	case XK_KP_Subtract:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing throttle or AT velocity");
		if ( aps_at_enabled(c) ) {
			aps_at_dec_velocity(c);
		}
		else {
			c->throttleComm -= 512;
			if ( c->throttleComm < 6553 )
				c->throttleComm = 6553;
		}
		break;

	case XK_3:
	case XK_KP_Add:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing throttle or AT velocity");
		if ( aps_at_enabled(c) ) {
			aps_at_inc_velocity(c);
		}
		else {
			c->throttleComm += 512;
			if ( c->throttleComm >= 32768 )
				c->throttleComm = 32768;
		}
		break;

	case XK_4:
/*	case XK_Prior:	*/

		fprintf(stderr, actfmt, curTime, symstr,
			"setting    throttle to full (disable all APS except AC)");
		aps_al_disable(c);
		aps_an_disable(c);
		aps_at_disable(c);
		aps_ap_disable(c);
		aps_aw_disable(c);

		c->throttleComm = 32768;
		break;
	
	case XK_exclam:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    pm_thrust_reverse_toggle");
		pm_thrust_reverse_toggle(c);
		break;

	case XK_a:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    pm_after_burner_toggle");
		pm_after_burner_toggle(c);
		break;

	case XK_space:
		/* Switch NAVn/RNAVn receivers: */
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    NAVs (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_switch_mode(u);
		break;

	case XK_7:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing OBS/ADF heading (disables AN)");
		aps_an_disable(c);
		hsi_obs_inc(u, -1);
		adf_hdg_inc(u, -1);
		break;

	case XK_8:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing OBS/ADF heading (disables AN)");
		aps_an_disable(c);
		hsi_obs_inc(u, +1);
		adf_hdg_inc(u, +1);
		break;

	case XK_9:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing HSI/ADF frequency (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_frq_inc(u, -1);
		adf_frq_inc(u, -1);
		break;

	case XK_0:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing HSI/ADF frequency (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_frq_inc(u, +1);
		adf_frq_inc(u, +1);
		break;

	case XK_F3:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing RNAV radial");
		hsi_theta_inc(u, -1);
		break;

	case XK_F4:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing RNAV radial");
		hsi_theta_inc(u, +1);
		break;

	case XK_F5:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing RNAV distance");
		hsi_rho_inc(u, -1);
		break;

	case XK_F6:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing RNAV distance");
		hsi_rho_inc(u, +1);
		break;

	case XK_b:
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   brakes");
		c->flags ^= FL_BRAKES;
		break;

	case XK_g:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    gear_handle_toggle");
		gear_handle_toggle(c);
		break;

	case XK_l:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    drone_new");
		drone_new(c, NULL);
		break;
		
/* Display next hint now with array wrapping */

	case XK_KP_1:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for next");
		prompt_viewer_print_hint(u, 1);
		break;

/* Display previous hint now with array wrapping */

	case XK_KP_7:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for previous");
		prompt_viewer_print_hint(u, 2);
		break;

/* Display first hint now */

	case XK_KP_9:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for first");
		prompt_viewer_print_hint(u, 3);
		break;

/* Display current hint now */

	case XK_KP_3:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for redisplay");
		prompt_viewer_print_hint(u, 0);
		break;

#ifdef sun
	case XK_Up:
#else
	case XK_KP_8:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for forward");
		setOutsideView(c, u, VIEW_FORWARD);
		prompt_viewer_print(u, "Front view");
		break;

/* look right */

#ifdef sun
	case XK_Right:
#else
	case XK_KP_6:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for right");
		setOutsideView(c, u, VIEW_RIGHT);
		prompt_viewer_print(u, "Right view");
		break;

/* look left */

#ifdef sun
	case XK_Left:
#else
	case XK_KP_4:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for left");
		setOutsideView(c, u, VIEW_LEFT);
		prompt_viewer_print(u, "Left view");
		break;

/* look back */

#ifdef sun
	case XK_Down:
#else
	case XK_KP_2:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for aft");
		setOutsideView(c, u, VIEW_AFT);
		prompt_viewer_print(u, "Rear view");
		break;

/* look up */

#ifdef sun
	case XK_R11:
#else
	case XK_KP_5:
#endif
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for up");
		setOutsideView(c, u, VIEW_UP);
		prompt_viewer_print(u, "Up view");
		break;

	case XK_KP_0:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for down");
		setOutsideView(c, u, VIEW_DOWN);
		prompt_viewer_print(u, "Down view");
		break;

	case XK_n:
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   chase/forward view");
		if( c->flags & FL_CHASE_VIEW ){
			setOutsideView(c, u, VIEW_FORWARD);
			prompt_viewer_print(u, "Front view");

		} else if( c->cinfo->object == NULL ){
			prompt_viewer_print(u, "Sorry, no object model available for this plane.");

		} else {
			setOutsideView(c, u, VIEW_CHASE);
			prompt_viewer_print(u, "Chase view");

		}
		break;

	case XK_q:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    newRadarTarget");
		c->curRadarTarget = newRadarTarget(c);
#ifdef HAVE_DIS
		dis_radarTargetChanged(c);
#endif
		break;

	case XK_D:
		/* Debug */
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    FSPageToggle");
		FSPageToggle();
		break;

	case XK_r:
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    radar without beam change (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);

		switch (c->radarMode) {

		case RM_OFF:
		case RM_STANDBY:
			if( u->c->cinfo->radarOutput > 0.0 ){
				/* we really have a radar -- enable it */
				setRadarMode(c, RM_NORMAL);
				enableRadar(c, RM_NORMAL);
			} else {
				/* no radar -- HSI mode */
				disableRadar(c);
				setRadarMode(c, RM_HSI);
				hsi_enable(u);
			}
			break;

		case RM_NORMAL:
		case RM_ACM:
		case RM_STT:
			disableRadar(c);
			setRadarMode(c, RM_HSI);
			hsi_enable(u);
			break;

		case RM_HSI:
			hsi_disable(u);
			setRadarMode(c, RM_ADF);
			adf_enable(u);
			break;

		case RM_ADF:
			adf_disable(u);
			setRadarMode(c, RM_OFF);
			break;

		default:
			adf_disable(u);
			disableRadar(c);
			setRadarMode(c, RM_OFF);
			hsi_enable(u);
			break;
		}

		break;

#ifdef HAVE_DIS
/* this is a hack to allow change of beam parameter index
in the outgoing electromagnetic emission pdu's */
	case XK_R:
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    radar with DIS beam change (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		switch (c->radarMode) {
		case RM_OFF:
		case RM_STANDBY:
			setRadarMode(c, RM_NORMAL);
			enableRadar(c, RM_NORMAL);
			break;
		case RM_NORMAL:
			disableRadar(c);
			setRadarMode(c, RM_ACM);
			enableRadar(c, RM_ACM);
			break;
		case RM_ACM:
			disableRadar(c);
			setRadarMode(c, RM_STT);
			enableRadar(c, RM_STT);
			break;
		case RM_STT:
			disableRadar(c);
			setRadarMode(c, RM_HSI);
			hsi_enable(u);
			break;
		case RM_HSI:
			disableRadar(c);
			hsi_disable(u);
			setRadarMode(c, RM_OFF);
			break;
		default:
			abort();
		}
		break;
#endif

#ifdef FIXME_INCOMPLETE_OR_BUGGED

	case XK_i:
		fprintf(stderr, actfmt, curTime, symstr,
			"saving     FLIGHT_STATE_FILE");
		fp = fopen(FLIGHT_STATE_FILE, "w");
		if ( fp == NULL ) {
			char s[1000];

			sprintf("Error opening the file " FLIGHT_STATE_FILE
				" for writing: %s", strerror(errno));
			prompt_craft_print(c, s);
		}
		else {
			fwrite((char *) &ptbl[c->pIndex],
				sizeof(craft), 1, fp);
			fclose(fp);
			prompt_craft_print(c, "Flight state saved in file "
				FLIGHT_STATE_FILE);
		}
		break;

	case XK_I:
		fprintf(stderr, actfmt, curTime, symstr,
			"restoring  FLIGHT_STATE_FILE");
		fp = fopen(FLIGHT_STATE_FILE, "r");
		if ( fp == NULL ) {
			char s[1000];

			sprintf("Error opening the file " FLIGHT_STATE_FILE
				" for reading: %s", strerror(errno));
			prompt_craft_print(c, s);
		}
		else {
			fread((char *) &pentry, sizeof(craft), 1, fp);
			fclose(fp);

			c->flags = pentry.flags;
			c->Sg = pentry.Sg;
			c->prevSg = pentry.prevSg;
			c->Cg = pentry.Cg;
			c->trihedral = pentry.trihedral;
			c->curHeading = pentry.curHeading;
			c->curPitch = pentry.curPitch;
			c->curRoll = pentry.curRoll;
			c->p = pentry.p;
			c->q = pentry.q;
			c->r = pentry.r;
			c->damageBits = pentry.damageBits;
			c->structurePts = pentry.structurePts;
			c->damageCL = pentry.damageCL;
			c->damageCM = pentry.damageCM;
			c->curThrust = pentry.curThrust;
			c->throttle = pentry.throttle;

			/* FIXME: there are missing fields */

			prompt_craft_print(c, "Flight state restored from file "
				FLIGHT_STATE_FILE);
		}
		break;

	case XK_semicolon:
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   debug");
		debug ^= 1;
		break;
#endif /* FIXME_INCOMPLETE_OR_BUGGED */

	//case XK_Escape:
	case XK_Q:
		if ( pressed_XK_Q_time + 2.0 >= curTime ){
			fprintf(stderr, actfmt, curTime, symstr,
				"executing  self-destruction");
			killPlayer(c, NULL);
			return TRUE;
		}
		else {
			fprintf(stderr, actfmt, curTime, symstr,
				"arming     self-destruction");
			pressed_XK_Q_time = curTime;
			prompt_viewer_print(u, "Press SHIFT-Q again within 2 sec. to exit the program");
		}
		break;

	case XK_P:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    system(xwd) for screen dump");
		system("xwd -name " ACM " -out /tmp/acm-dump-`date +%s`.xwd");
		prompt_viewer_print(u, "Screen dumped into /tmp/acm-dump-*.xwd");
		break;

#ifdef FIXME_INCOMPLETE_OR_BUGGED

	case XK_braceleft:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    startBlackBoxRecording");
		startBlackBoxRecording();
		prompt_viewer_print(u, "Black box recording started");
		break;

	case XK_braceright:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    endBlackBoxRecording");
		endBlackBoxRecording();
		prompt_viewer_print(u, "Black box recording stopped");
		break;

	case XK_bracketleft:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    startBlackBoxPlayback");
		startBlackBoxPlayback();
		prompt_viewer_print(u, "Black box playback started");
		break;

#endif

	case XK_F7:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing altimeter correction");
		inst_altimeter_correction(u, -1);
		break;

	case XK_F8:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing altimeter correction");
		inst_altimeter_correction(u, +1);
		break;

	case XK_F9:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    inst_attitude_reset");
		inst_attitude_reset(u);
		break;

	case XK_F11:
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing attitude pitch");
		inst_attitude_adjust_pitch(u, -DEGtoRAD(0.5));
		break;

	case XK_F12:
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing attitude pitch");
		inst_attitude_adjust_pitch(u, DEGtoRAD(0.5));
		break;

	case XK_plus:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    windows_zoom_in");
		windows_zoom_in(u);
		break;

	case XK_minus:
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    windows_zoom_out");
		windows_zoom_out(u);
		break;

	case XK_dollar:
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   drone commands mode");
		if( c->type == CT_PLANE )
			drone_take_commands(c);
		else
			drone_release_commands(c);
		break;

/* Handle without doing anything */

	case XK_Caps_Lock:
	case XK_Shift_Lock:
	case XK_Num_Lock:
	case XK_Shift_L:
	case XK_Shift_R:
	case XK_Control_L:
	case XK_Control_R:
	case XK_Alt_L:
	case XK_Alt_R:
	case XK_Meta_L:
	case XK_Meta_R:
	case XK_Super_L:
	case XK_Super_R:
	case XK_Hyper_L:
	case XK_Hyper_R:
		fprintf(stderr, actfmt, curTime, symstr,
			"ignoring   silently");
		break;

/* Currently unassigned in events.c */

/*	case XK_5:	*/
/*	case XK_6:	*/
	case XK_E:
	case XK_F:
	case XK_J:
	case XK_O:
	case XK_S:
	case XK_W:
	case XK_Y:
	case XK_e:
	case XK_f:
	case XK_k:
	case XK_o:		/* used by browse */
	case XK_p:		/* used by browse */
	case XK_u:
	case XK_KP_Begin:
	case XK_KP_Decimal:
	case XK_KP_Delete:
	case XK_KP_Divide:
	case XK_KP_Down:
	case XK_KP_End:
	case XK_KP_Enter:
	case XK_KP_Home:
	case XK_KP_Insert:
	case XK_KP_Left:
	case XK_KP_Multiply:
	case XK_KP_Next:		/* used by browse */
	case XK_KP_Prior:		/* used by browse */
	case XK_KP_Right:
	case XK_KP_Up:
	case XK_F2:
	case XK_F10:
	case XK_BackSpace:
	case XK_Delete:
	case XK_Escape:
	case XK_Insert:
	case XK_Menu:
	case XK_Pause:
	case XK_Print:
	case XK_Return:
	case XK_Scroll_Lock:
	case XK_ampersand:
	case XK_apostrophe:
/*	case XK_asciicircum:	*/
	case XK_asciitilde:
	case XK_asterisk:
	case XK_at:
	case XK_backslash:
	case XK_bracketright:
	case XK_colon:
	case XK_equal:
	case XK_grave:
	case XK_numbersign:
/*	case XK_percent:	*/
	case XK_question:
	case XK_quotedbl:
	case XK_underscore:
	default:

/* Unhandled input (probably typo) -- warn pilot */

		fprintf(stderr, actfmt, curTime, symstr,
			"warning    pilot about probable typo\a");
		sprintf(viewermsg, "%16s is not a supported input!", symstr);
		prompt_viewer_print(u, viewermsg);
		break;
	}

	return FALSE;
}


static _BOOL
doWindowEvent(viewer * u)
/*
	Return TRUE on:
	
	- close request of the window: in this case the viewer u gets closed
	  and released
	
	- player gets killed or terminated: in this case the viewer u and all
	  the other viewers attached to the craft c gets closed and the craft c
	  gets released
*/
{
	XEvent    ev;
	craft    *c;

	updateSystemClock();
	fprintf(stderr, "%s\n", "=> doWindowEvent");
	c = u->c;

/*	updateSystemClock();	*/
/*	fprintf(stderr, "%s\n", "into  XCheckWindowEvent");	*/
	if (XCheckWindowEvent(u->dpy, u->win, -1L, &ev)) {
/*		updateSystemClock();	*/
/*		fprintf(stderr, "%s\n", "outof XCheckWindowEvent--true");	*/

		switch (ev.type) {

		case MappingNotify:
			XRefreshKeyboardMapping(&ev.xmapping);
			break;

		case KeyPress:
			if (u->viewer_state == ViewerStateBrowsing ||
				u->viewer_state == ViewerStatePiggyback) {
				if (doBrowseKeyEvent( c, u, &ev, 1)) {
					return FALSE;
				}
			}
			else {
				if ( doKeyEvent(c, u, (XKeyEvent *) &ev) ){
					return TRUE;
				}
			}
			break;

		case KeyRelease:
			if ( doKeyRelease(c, u, (XKeyEvent *) &ev) ){
				return TRUE;
			}
			break;

		case ButtonPress:
			if (doButtonEvent(c, u, &ev))
				return TRUE;
			break;

		case ButtonRelease:
			if (doButtonReleaseEvent(c, u, &ev, 1))
				return TRUE;
			break;

		case ConfigureNotify:
			windows_set_layout(c, u,
				ev.xconfigure.width,
				ev.xconfigure.height, u->hud_mode);
			break;

		case FocusIn:
			u->hasFocus = TRUE;
			break;

		case FocusOut:
			u->hasFocus = FALSE;
			u->hasComm = FALSE;
			break;

		case Expose:
			ForceWindowRedraw(u->w);
			break;

		default:
			break;
		}

	} else {
/*		updateSystemClock();	*/
/*		fprintf(stderr, "%s\n", "outof XCheckWindowEvent--false");	*/
/*		fprintf(stderr, "%s\n", "into  XCheckTypedEvent--ClientMessage");	*/
		if (XCheckTypedEvent(u->dpy, ClientMessage, &ev)) {
			updateSystemClock();
			fprintf(stderr, "%s\n", "outof XCheckTypedEvent--ClientMessage--true");

			if (ev.xclient.message_type == u->protocolsAtom
			&& (ev.xclient.data.l[0] == u->deleteWindowAtom
			|| ev.xclient.data.l[0] == u->closedownAtom)) {
/*				updateSystemClock();	*/
/*				fprintf(stderr, "%s\n", "into  killPlayer");	*/
				killPlayer(c, NULL);
/*				updateSystemClock();	*/
/*				fprintf(stderr, "%s\n", "outof killPlayer");	*/
				return TRUE;
			}
		} else {
/*			updateSystemClock();	*/
/*			fprintf(stderr, "%s\n", "outof XCheckTypedEvent--ClientMessage--false");	*/
/*			fprintf(stderr, "%s\n", "into  XCheckTypedEvent--MappingNotify");	*/
			if (XCheckTypedEvent(u->dpy, MappingNotify, &ev)) {
				updateSystemClock();
				fprintf(stderr, "%s\n", "outof XCheckTypedEvent--MappingNotify--true");
/*				fprintf(stderr, "%s\n", "into  XRefreshKeyboardMapping");	*/
				XRefreshKeyboardMapping(&ev.xmapping);
				updateSystemClock();
				fprintf(stderr, "%s\n", "outof XRefreshKeyboardMapping");
				return TRUE;
			} else {
/*				updateSystemClock();	*/
/*				fprintf(stderr, "%s\n", "outof XCheckTypedEvent--MappingNotify--false");	*/
/*				fprintf(stderr, "%s\n", "into  XCheckTypedEvent--SelectionClear");	*/
				if (XCheckTypedEvent(u->dpy, SelectionClear, &ev)) {
					updateSystemClock();
					fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionClear--true");
				} else {
/*					updateSystemClock();	*/
/*					fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionClear--false");	*/
/*					fprintf(stderr, "%s\n", "into  XCheckTypedEvent--SelectionNotify");	*/
					if (XCheckTypedEvent(u->dpy, SelectionNotify, &ev)) {
						updateSystemClock();
						fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionNotify--true");
					} else {
/*						updateSystemClock();	*/
/*						fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionNotify--false");	*/
/*						fprintf(stderr, "%s\n", "into  XCheckTypedEvent--SelectionRequest");	*/
						if (XCheckTypedEvent(u->dpy, SelectionRequest, &ev)) {
							updateSystemClock();
							fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionRequest--true");
						} else {
							updateSystemClock();
							fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionRequest--false");
						}
					}
				}
			}
		}
	}

/*	updateSystemClock();	*/
/*	fprintf(stderr, "%s\n", "<= doWindowEvent returning false (no events in queue)");	*/
	return FALSE;  /* done with all pending events for this viewer */

}


void
events_window_keyb_buttons(void)
/*
	Process events for each viewer
*/
{
	viewer   *u;
	_BOOL     got_evt;

#ifdef FIXME
	int n;

	n = 0;
	u = vl_head;
	while( u != NULL ){
		n++;
		u = u->vl_next;
	}
	printf("FIXME: trovati %d viewers\n", n);
#endif

	do {

		/*
			Process one event for each viewer
		*/

		got_evt = FALSE;

		u = vl_head;

		while( u != NULL ){

			if (real_delta_t || ((cur % redraw_interval) == 0)) {
				getStick(u->c, u, 0);
			}

			if( XEventsQueued(u->dpy, QueuedAlready) ){

				got_evt = TRUE;

				if( doWindowEvent(u) ){

					/*
					Viewer or craft terminated. We can't
					continue scanning the linked list of
					viewers because we don't know how many
					viewers were attached to this craft, so
					we have to restart from the first
					viewer:
					*/

					u = vl_head;

				} else {

					u = u->vl_next;
				}

			} else {

				u = u->vl_next;
			}
		}
	
	} while( got_evt );
}


static int last_switches = 0;


int
events_joystick(craft * c, viewer * u, double throttle, int switches)
{

	int       switch_xor = switches ^ last_switches;

/*
 *  Change in state of any buttons ?
 */

	if (switch_xor != 0) {

		if (switch_xor & 1) {

/*
 *  Button 1 press
 */
			if (switches & 1) {
				fireWeapon(c);
			}

/*
 *  Button 1 release
 */

			else {
				ceaseFireWeapon(c);
			}
		}

/*
 * Button 2 press
 */

		if (switch_xor & 2) {
			if (switches & 2) {
				selectWeapon(c);
			}
		}

	}

	last_switches = switches;

	return 0;
}

